import {ParsedConfiguration, readConfiguration} from '@angular/compiler-cli'

import * as ts from 'typescript'

import {ImportDependency, PathTransform} from 'parcel-plugin-typescript/exports'

import {CompileResult} from '../../interfaces'
import {AngularCompilerHost} from './host'
import {generateRouteLoader} from './route'

import {BuildLoop} from './build-loop'

const COLLECT_DEPS = true

export class AngularCompiler {
	public readonly host: AngularCompilerHost
	public readonly loop: BuildLoop

	private readonly config: ParsedConfiguration
	private readonly dependencies = COLLECT_DEPS ? new Map<string, ImportDependency[]>() : null
	private readonly transformers: Array<ts.TransformerFactory<ts.SourceFile>> = []

	constructor(project: string, compileResource: (file: string) => Promise<string>) {
		this.config = readConfiguration(project)

		const {options} = this.config

		this.host = new AngularCompilerHost(options, compileResource)
		this.loop = new BuildLoop(this.host, this.config, this.transformers)

		this.transformers.push(PathTransform(options, this.dependencies))
	}

	/**
	 * Triggers a TypeScript compilation if needed and returns a compiled file
	 * @param path a absolute path to a TypeScript source file to compile
	 */
	public async compile(path: string): Promise<CompileResult<string[]>> {
		const {config, dependencies, host, loop} = this

		if(loop.entryFile === null) {
			// We assume the file file included by the project is the entry
			// It is used to inject the generated SystemJS loader
			// TODO: use Parcel dependencies instead
			loop.entryFile = path
		}

		await loop.emit(undefined)

		const {basePath, outDir} = config.options

		if(basePath && outDir) {
			path = path.replace(basePath, outDir)
		}

		const program = loop.getProgram()
		let js = host.store.readFile(path.replace(/\.tsx?$/, '.js'))!

		// detect if the file is the main module
		// TODO: use Parcel dependencies
		if(program && path === loop.entryFile) {
			js = `${js}\n${generateRouteLoader(path, program)}`
		}

		return {
			sources: {js},
			resources: loop.getResources(path),
			dependencies: dependencies
				? dependencies.get(path) || []
				: null
		}
	}
}
